package org.barney.greenfoot.world;
import greenfoot.*;
import java.util.*;
/**
 * Objekte dieser Klasse generieren Zufallswelten mit angegebenen Klassen und Wahrscheinlichkeiten.
 * Um eine Welt generieren zu knnen, muss eine Instanz erzeugt werden, der man 
 * eine <code>ArrayList</code> bergibt. Diese enthlt die zu generierenden Objektklassen und 
 * die Spawnwahrscheinlichkeiten. Mit der Methode <code>generate(World, float, boolean, boolean)</code> werden nun 
 * in der angegebenen Welt Objekte generiert.
 * <p>
 * Die Spawnwahrscheinlichkeiten sind relativ zueinander. Z.B. liegt bei den Wahrscheinlichkeiten 
 * 15f und 45f ein 1:3-Verhltnis vor. <code>Generatable</code>-Objekte bei denen <code>isStatic == true</code> ist, 
 * werden vor den anderen Objekten erzeugt. Ebenso ist es mglich, dass mehr Objekte spawnen als erwnscht, 
 * da der Generator so oder so soviele statische Objekte spawnt, wie es in den <code>Generatable</code>-Objekten vermerkt ist.
 * 
 * @author  Jan-Niklas Brandes 
 * @version  1.0
 */
public class RandomWorldGenerator
{
    /**
     * Die Liste mit den zu spawnenden <code>Generatable</code>-Objekten.
     */
    protected ArrayList<Generatable> generatorObjects;
    /**
     * Instantiiert einen <code>RandomWorldGenerator</code>.
     * @param  classes Die <code>Generatable</code>-Objekte, die die Spawnklassen enthalten.
     */
    public RandomWorldGenerator(ArrayList<Generatable> classes){
        generatorObjects = classes;
    }
    /**
     * Generiert in der angegebenen Welt zufllige Strukturen. 
     * Vorherige Strukturen werden gelscht.
     * @param  world Die Welt, in der generiert werden soll.
     * @param  surfacePercentage Wie viel Prozent der Flche genutzt werden soll (z.B. <code>0.4f</code>)
     * @param  surfacePercentageAsObjectNumber Ob <code>surfacePercentage</code> eine feste Anzahl an Spawnobjekten sein soll.
     * @param  staticCounts Ob statische Objekte in der <code>surfacePercentage</code> miteinbezogen werden.
     */
    public void generate(World world, float surfacePercentage, boolean surfacePercentageAsObjectNumber, boolean staticCounts){
        world.removeObjects(new ArrayList(world.getObjects(Actor.class)));
        float objectNumber = 0f;
        if(surfacePercentageAsObjectNumber){
            objectNumber = surfacePercentage;
        }else{
            objectNumber = (world.getWidth()*world.getHeight())*(surfacePercentage);
        }
        float allProb = 0f;
        for(Generatable g : generatorObjects){
            if(!g.isStatic){
                allProb += g.probability;
            }else{
                for(int i = 0; i<g.probability; i++){
                    int rndX = 0;
                    int rndY = 0;
                    do{
                        Coordinate rndCo = generateRandomCoordinate(world);
                        rndX = rndCo.getX();
                        rndY = rndCo.getY();
                    }while(!maySpawn(world, g, rndX, rndY));
                    try{
                        world.addObject(g.generatable.newInstance(), rndX, rndY);
                        objectNumber--;
                    }catch(Exception ex){ex.printStackTrace();}
                }
            }
        }
        for(Generatable g : generatorObjects){
            if(!g.isStatic){
                int alreadyInWorld = world.getObjects(g.generatable).size();
                int remainingObjects = (int)objectNumber*((int)(g.probability/allProb));
                for(int i = 0; i<(staticCounts ? (int)objectNumber*(g.probability/allProb) : ((int)objectNumber*(g.probability/allProb))-alreadyInWorld); i++){
                    int rndX = 0;
                    int rndY = 0;
                    do{
                        Coordinate rndCo = generateRandomCoordinate(world);
                        rndX = rndCo.getX();
                        rndY = rndCo.getY();
                    }while(!maySpawn(world, g, rndX, rndY));
                    try{
                        world.addObject(g.generatable.newInstance(), rndX, rndY);
                    }catch(Exception ex){ex.printStackTrace();}
                }
            }
        }
    }
    /**
     * Platziert ein zuflliges <code>Generatable</code> aus der Liste in der Welt <code>world</code>.
     * @param  world Die Welt, in der gespawnt wird.
     * @param  x Der x-Wert der Spawnkoordinate.
     * @param  y Der y-Wert der Spawnkoordinate.
     */
    public void placeRandomObject(World world, int x, int y){
        int rndIndex = Greenfoot.getRandomNumber(generatorObjects.size());
        Actor g = null;
        try{
            g = (Actor)generatorObjects.toArray(new Generatable[1])[rndIndex].generatable.newInstance();
        }catch(Exception ex){ex.printStackTrace();}
        if(x < world.getWidth() && x >= 0 && y < world.getHeight() && y >= 0 && maySpawn(world, generatorObjects.toArray(new Generatable[1])[rndIndex], x, y)){
            world.addObject(g, x, y);
        }
    }
    /**
     * Generiert eine Zufallskoordinate in den Grenzen der angegebenen Welt.
     * @param  world Die Welt, deren Gre bercksichtigt werden soll.
     * @return  Eine Zufallskoordinate.
     */
    public Coordinate generateRandomCoordinate(World world){
        return new Coordinate(Greenfoot.getRandomNumber(world.getWidth()), Greenfoot.getRandomNumber(world.getHeight()));
    }
    /**
     * Gibt zurck, ob das angegebene <code>Generatable</code> an angegebener Stelle spawnen kann.
     * @param  world Die Welt, in der gespawnt werden wrde.
     * @param  g Das Generatable, dass gespawnt werden soll.
     * @param  x Der x-Wert der potentiellen Spawnkoordinate.
     * @param  y Der y-Wert der potentiellen Spawnkoordinate.
     * @return  Ob der Spawnplatz geeignet ist.
     */
    public boolean maySpawn(World world, Generatable g, int x, int y){
        return world.getObjectsAt(x, y, Actor.class).size() == 0;
    }
}
